#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "sha1.h"
#include "logger.h"
#include "accounting.h"
#include "kv_list.h"

kv_list *userlist = NULL;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

static int _load_userlist(void)
{
	int ret;
	char *login = NULL;
	struct usr_t user;

	FILE *fd = fopen("users.lst", "rb");
	if (fd == NULL)
	{
		return 0;
	}

	while (1)
	{
		ret = fread(&user, sizeof(struct usr_t), 1, fd);
		if (ret != 1) break;
		login = malloc(strlen(user.username) + 1);
		memcpy(login, user.username, strlen(user.username)+1);
		ret = _put_user(login, &user);
		if (ret != 0)
		{
			logger(LOG_WARN, "[M] Loading userlist, put_user(\"%s\"): %d", login, ret);
		}
	}

	fclose(fd);
	return 0;
}

static int _save_userlist(void)
{
	kv_list *list = userlist;
	if (list == NULL) return 0;

	pthread_mutex_lock(&mutex);
	FILE *fd = fopen("users.lst", "wb");
	if (fd == NULL)
	{
		pthread_mutex_unlock(&mutex);
		return -1;
	}

	while(1)
	{
		fwrite(list->value, sizeof(struct usr_t), 1, fd);
		if (list->next == NULL) break;
		list = list->next;
	}

	fflush(fd);
	fclose(fd);
	pthread_mutex_unlock(&mutex);
	return 0;
}

int check_session(char *sess1, char *sess2)
{
	return memcmp(sess1, sess2, SESSION_LEN);
}

int generate_session(char *session, char *login, char *password)
{
	/*
	*  1: hash = sha1(login + password + time)
	*  2: for example hash == 5b 91 a5 f6 69 bf 08 8d 9d 3b 79 7c 40 06 98 fb 44 f0 c7 b9
	*                         0  1  2  3  4  5  6  7  8  9  0  1  2  3  4  5  6  7  8  9
	*     session[0, 1] = 5b + 70
	*     session[2, 3] = 91 + 7c
	*     ...
	*     session[18, 19] = 3b + b9
	*/

	time_t currtime;
	time(&currtime);

	char data[LOGIN_MAX_LEN + PASS_MAX_LEN + TIMESTR_LEN + 1];
	char hash[21] = { 0 };

	sprintf(data, "%s%s%u", login, password, (unsigned)currtime);
	sha1(hash, data);

	for (int i=0; i<10; i++)
	{
		sprintf(session+i*2, "%02x", (unsigned char)(hash[i] + hash[i+10]));
	}
	return 0;
}

int _db_init(void)
{
	return _load_userlist();
}

int _db_cleanup(void)
{
	return _save_userlist();
}

int _get_user(char *login, struct usr_t *user)
{
	kv_list *list = userlist;
	if (list == NULL)
	{
		return -1;
	}

	while (1)
	{
		if (strcmp(list->key, login) == 0)
		{
			memcpy(user, list->value, sizeof(struct usr_t));
			return 0;
		}
		if (list->next == NULL) break;
		list = list->next;
	}
	return -1;
}

int _put_user(char *login, struct usr_t *user)
{
	int ret;
	static int counter = 0;
	kv_list *list = NULL;
	kv_list *found = NULL;
	size_t key_size = strlen(login) + 1;
	size_t val_size = sizeof(struct usr_t);

	counter++;
	if (counter == 10)
	{
		ret = _save_userlist();
		if (ret != 0)
		{
			logger(LOG_WARN, "[M] Saving userlist: %d", ret);
		}
		counter = 0;
	}

	if (userlist == NULL)
	{
		pthread_mutex_lock(&mutex);

		userlist = malloc(sizeof(kv_list));

		userlist->key = malloc(key_size);
		memcpy(userlist->key, login, key_size);

		userlist->value = malloc(val_size);
		memcpy(userlist->value, user, val_size);

		userlist->next = NULL;

		pthread_mutex_unlock(&mutex);
		return 0;
	}

	list = userlist;
	pthread_mutex_lock(&mutex);
	while (1)
	{
		if (strcmp(list->key, login) == 0)
		{
			found = list;
			break;
		}
		if (list->next == NULL) break;
		list = list->next;
	}

	if (found == NULL)
	{
		list->next = malloc(sizeof(kv_list));
		list = list->next;
	}
	else
	{
		free(list->key);
		free(list->value);
	}

	list->key = malloc(key_size);
	memcpy(list->key, login, key_size);

	list->value = malloc(val_size);
	memcpy(list->value, user, val_size);

	if (found == NULL)
	{
		list->next = NULL;
	}

	pthread_mutex_unlock(&mutex);
	return 0;
}

int _del_user(char *login)
{
	kv_list *list = userlist;
	kv_list *prev = NULL;
	if (list == NULL) return -1;

	/* Checking first element this way, as we'll not be able to change userlist later */
	if (strcmp(userlist->key, login) == 0)
	{
		pthread_mutex_lock(&mutex);
		free(userlist->key);
		free(userlist->value);

		prev = userlist;
		userlist = userlist->next;

		free(prev);
		pthread_mutex_unlock(&mutex);
		return 0;
	}

	prev = list;

	while (1)
	{
		if (strcmp(list->key, login) == 0)
		{
			pthread_mutex_lock(&mutex);
			free(list->key);
			free(list->value);
			prev->next = list->next;
			free(list);
			pthread_mutex_unlock(&mutex);
			return 0;
		}
		if (list->next == NULL) break;
		prev = list;
		list = list->next;
	}
	return -1;
}