#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <dirent.h>
#include <dlfcn.h>
#include <unistd.h>

#include "networking.h"
#include "messages.h"
#include "core.h"
#include "accounting.h"
#include "packets.h"
#include "logger.h"
#include "sha1.h"
#include "modules.h"
#include "datatypes.h"
#include "kv_list.h"

/*
*  TODO: get rid of var_table in favor of some kind of API
*		`gpar_t global` looks awful
*/
kv_list			*var_table;
kv_list			*config;
struct gpar_t		global;
kv_list			*modules;
kv_list			*custom_packets;

int (*db_init)(void);
int (*db_cleanup)(void);

int (*get_user)(char*, struct usr_t*);
int (*put_user)(char*, struct usr_t*);
int (*del_user)(char*);

int main(int argc, char **argv)
{
	MC_Init(argc, argv);
	MC_RunModules(MASTER_START, NULL);

	if (MC_RunThread(server_thread, NULL) != 0)
	{
		exit(EXIT_FAILURE);
	}

	while (global.exit_flag == 0)
	{
		sleep(10);
	}

	MC_Cleanup();
	return EXIT_SUCCESS;
}

void *server_thread(void *param)
{
	/* shut up, gcc :< */
	param = NULL;
	/* _______________ */

	int client_sock;
	struct sockaddr_in client_addr;
	socklen_t cli_len = sizeof(client_addr);

	while (1)
	{
		client_sock = accept(global.socket, (struct sockaddr *)&client_addr, &cli_len);
		if (client_sock < 0)
		{
			if (global.exit_flag == 1)
			{
				pthread_exit(NULL);
			}
			continue;
		}

		struct client_t *cli_info = malloc(sizeof(struct client_t));
		cli_info->sock = client_sock;
		cli_info->addr = client_addr.sin_addr;

		MC_TuneSocket(client_sock);

		list_add(&var_table, "client_sock", &client_sock, 0);
		list_add(&var_table, "client_addr", &client_addr, 0);
		list_add(&var_table, "client_info", cli_info, FL_FREE);

		MC_RunModules(NEW_CONNECTION, NULL);
		MC_RunThread(client_thread, (void*)cli_info);
	}
}

void *client_thread(void *param)
{
	struct client_t *client = param;
	char main_buf[RECV_MAX] = { 0 };
	ssize_t buf_size;
	int req_type;

	buf_size = MC_RecvAll(client->sock, main_buf, RECV_MAX);
	if (buf_size == 0)
	{
		close(client->sock);
		pthread_exit(NULL);
	}
	req_type = get_packet_type(main_buf);

	/* widely used variables */
	struct usr_t *curr_user = malloc(sizeof(struct usr_t));
	char answer[256] = { 0 };

	/* "login" is "username" with all characters lowered (a-z) */
	char login[LOGIN_MAX_LEN+1];
	char username[LOGIN_MAX_LEN+1];
	char password[PASS_MAX_LEN+1] = { 0 };
	char enc_password[PASS_MAX_LEN+1] = { 0 };
	char session[SESSION_LEN+1] = { 0 };
	/* ______________________ */

	list_add(&var_table, "main_buf", main_buf, 0);
	list_add(&var_table, "user_login", login, 0);
	list_add(&var_table, "user_password", enc_password, 0);
	list_add(&var_table, "user_session", session, 0);
	list_add(&var_table, "user_info", curr_user, FL_FREE);

	 /**********************************************
	 * SOME COMMENTS MAY BE INCORRECT AND OUTDATED *
	 **********************************************/

	switch(req_type)
	{
	case TYPE_GETVERSION:
	{
		/****************************************************************************************
		*  While parsing main_buf, we are looking for 3 fields: login, password and version
		*  If there're 0-2 scanned fields reported by sscanf, then something is going wrong
		*  Login and password should be 4-20 chars long and version should match LAUNCHER_VERSION
		****************************************************************************************/
		char version[3] = { 0 };
		int ret;

		ret = sscanf(main_buf, "/game/getversion.jsp\nuser=%20[^&]&password=%20[^&]&version=%2s", username, password, version);

		if ((ret < 3) || (strlen(username) < 4) || (strlen(password) < 4))
		{
			sendstr(client->sock, LOGIN_PASS_RULES);
			break;
		}

		if (strcmp(version, global.launcher_version) != 0)
		{
			sendstr(client->sock, OUTDATED_LAUNCHER);
			break;
		}

		strtolower(login, username);
		ret = get_user(login, curr_user);
		if (ret != 0)
		{
			sendstr(client->sock, WRONG_LOGIN_PASS);
			break;
		}

		sha1(enc_password, password);
		if (memcmp(enc_password, curr_user->password, PASS_MAX_LEN) != 0)
		{
			sendstr(client->sock, WRONG_LOGIN_PASS);
			break;
		}

		generate_session(session, login, enc_password);
		snprintf(answer, 255, "%s:%s:%s:%s:%s:", global.game_version, global.ticket, curr_user->username, session, global.launcher_version);

		memcpy(curr_user->session, session, SESSION_LEN+1);
		curr_user->lastlogin = time(NULL);

		ret = put_user(login, curr_user);
		if (ret != 0)
		{
			sendstr(client->sock, UNKNOWN_ERROR);
			break;
		}

		sendstr(client->sock, answer);

		break;
	}
	case TYPE_JOINSERVER:
	{
		/****************************************************************************************
		*  We are looking for 2 fields now: login and session
		*  If there're 0-1 scanned fields reported by sscanf, then something is going wrong
		****************************************************************************************/
		char *allow = "HTTP/1.0 200 OK\n\nOK";
		char *disallow = "You are not allowed to join this server!";
		int ret;

		ret = sscanf(main_buf, "GET /game/joinserver.jsp?user=%20[^&]&sessionId=%20[^&]&%*s", username, session);
		if (ret < 2)
		{
			sendstr(client->sock, disallow);
			break;
		}

		strtolower(login, username);
		ret = get_user(login, curr_user);
		if (ret != 0)
		{
			sendstr(client->sock, disallow);
			break;
		}

		if (check_session(curr_user->session, session) != 0)
		{
			sendstr(client->sock, disallow);
			break;
		}

		sendstr(client->sock, allow);

		break;
	}
	case TYPE_CHECKSERVER:
	{
		char *allow = "HTTP/1.0 200 OK\n\nYES";
		int res;

		res = sscanf(main_buf, "GET /game/checkserver.jsp?user=%20[^&]&%*s", username);
		if (res < 1)
		{
			sendstr(client->sock, UNKNOWN_ERROR);
			break;
		}

		sendstr(client->sock, allow);
		break;
	}
	case TYPE_FILES:
	{
		/* Filename_full should have enough space for 'files/' and requested file name */
		char filename[32+1] = { 0 };
		char filename_full[32+6+1] = { 0 };
		int res;

		res = sscanf(main_buf, "GET /game/%32[^? ]", filename);
		if (res < 1)
		{
			sendstr(client->sock, UNKNOWN_ERROR);
			break;
		}

		sprintf(filename_full, "files/%s", filename);
		MC_Sendfile(client->sock, filename_full, NULL);
		break;
	}
	case TYPE_REGISTER:
	{
		int version = 0;
		int ret;

		ret = sscanf(main_buf, "/game/register.jsp\nuser=%20[^&]&password=%20[^&]&version=%d", username, password, &version);
		if ((ret < 3) || (strlen(username) < 5) || (strlen(password) < 5))
		{
			sendstr(client->sock, LOGIN_PASS_RULES);
			break;
		}

		strtolower(login, username);
		ret = get_user(login, curr_user);
		if (ret == 0)
		{
			sendstr(client->sock, USERNAME_TAKEN);
			break;
		}

		curr_user->regtime = time(NULL);
		memcpy(curr_user->username, username, LOGIN_MAX_LEN+1);

		sha1(enc_password, password);
		memcpy(curr_user->password, enc_password, PASS_MAX_LEN+1);

		memset(curr_user->session, 0, SESSION_LEN+1);
		curr_user->flags = 0;

		ret = put_user(login, curr_user);
		if (ret != 0)
		{
			sendstr(client->sock, UNKNOWN_ERROR);
			break;
		}

		sendstr(client->sock, "Registered successfully!");
		break;
	}
	default:
		MC_RunModules(CUSTOM_PACKET, main_buf);
		break;
	}

	close(client->sock);
	pthread_exit(NULL);
}

void parse_config(char *filename)
{
	FILE *cfg = NULL;
	char conf_key[64+1];
	char conf_value[64+1];
	void *p_value = NULL;
	int ret;

	config = NULL;

	cfg = fopen(filename, "rb");
	if (cfg == NULL)
	{
		logger(LOG_INFO, "Can't open config file: %s, I'll create one for you", filename);
		cfg = fopen(filename, "wb");
		if (cfg == NULL)
		{
			return;
		}
		fclose(cfg);
		return;
	}

	while (ret != EOF)
	{
		ret = fscanf(cfg, "%64[^ \n] = %64[^ \n]\n", conf_key, conf_value);
		if (ret == 2)
		{
			if (conf_key[0] == '#') continue;

			p_value = malloc(strlen(conf_value) + 1);
			strcpy(p_value, conf_value);

			list_add(&config, conf_key, p_value, FL_FREE);
		}
	}
}

/* TODO: separate signal handling and other shit */
void MC_Init(int argc, char **argv)
{
	/* later argc and argv will be used properly */
	argc = 0;
	argv = NULL;
	/* _________________________________________ */

	int ret;

	config = NULL;
	parse_config("master.conf");
	logger(LOG_INFO, "Config file \"%s\" parsed successfully", "master.conf");

	global.log_level = strtol(list_get(config, "log_level"), NULL, 10);
	if (global.log_level > 4 || global.log_level < 0)
	{
		logger(LOG_WARN, "Logging level from \"%s\" is out of possible range [0-4]. I'll set it to 0.", "master.conf");
		global.log_level = 0;
	}

	/* Setting ignore action on SIGPIPE */
	struct sigaction act;
	act.sa_handler = SIG_IGN;
	ret = sigaction(SIGPIPE, &act, NULL);
	if (ret == -1)
	{
		logger(LOG_CRIT, "Couldn't set SIGPIPE handler");
		exit(EXIT_FAILURE);
	}
	logger(LOG_DEBUG, "SIGPIPE handler was set to SIG_IGN @ %0.8p", SIG_IGN);

	/* Setting ignore action on SIGINT */
	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler = &signal_handler;
	ret = sigaction(SIGINT, &act, NULL);
	if (ret == -1)
	{
		logger(LOG_CRIT, "Couldn't set SIGINT handler");
		exit(EXIT_FAILURE);
	}
	logger(LOG_DEBUG, "SIGINT handler was set to \"signal_handler\" @ %0.8p", signal_handler);

	/* Setting ignore action on SIGTERM (default in `kill`) */
	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler = &signal_handler;
	ret = sigaction(SIGTERM, &act, NULL);
	if (ret == -1)
	{
		logger(LOG_CRIT, "Couldn't set SIGTERM handler");
		exit(EXIT_FAILURE);
	}
	logger(LOG_DEBUG, "SIGTERM handler was set to \"signal handler\" @ %0.8p", signal_handler);

	/* setting necessary options to defaults */
	global.exit_flag = 0;
	if ((global.game_version = list_get(config, "game_version")) == NULL)
	{
		global.game_version = DEFAULT_GAME_VERSION;
	}
	if ((global.launcher_version = list_get(config, "launcher_version")) == NULL)
	{
		global.launcher_version = DEFAULT_LAUNCHER_VERSION;
	}
	if ((global.ticket = list_get(config, "main_ticket")) == NULL)
	{
		global.ticket = DEFAULT_TICKET;
	}
	logger(LOG_DEBUG, "Essential variables were successfully set:\n  game_version = \"%s\"\n  launcher_version = \"%s\"\n  main_ticket = \"%s\"", global.game_version, global.launcher_version, global.ticket);

	/* Database functions */
	db_init = _db_init;
	db_cleanup = _db_cleanup;

	get_user = _get_user;
	put_user = _put_user;
	del_user = _del_user;

	char *database_lib = list_get(config, "database_lib");
	if (database_lib != NULL)
	{
		init_database_lib(database_lib, &db_init, &db_cleanup, &get_user, &put_user, &del_user);
	}
	logger(LOG_DEBUG, "DB functions:\n  db_init @ %p\n  db_cleanup @ %p\n  get_user @ %p\n  put_user @ %p\n  del_user @ %p", db_init, db_cleanup, get_user, put_user, del_user);

	ret = db_init();
	if (ret != 0)
	{
		logger(LOG_ERROR, "db_init() returned %d, database module may behave incorrectly", ret);
	}

	/* TODO: MC_InitModules() must return something. MORE CHECKS! MORE CODE! */
	/* Modules */
	modules = NULL;
	custom_packets = NULL;
	var_table = NULL;

	MC_InitModules();

	global.socket = MC_PrepareSocket();
	if (global.socket < 0)
	{
		exit(EXIT_FAILURE);
	}

	list_add(&var_table, "modules", modules, 0);
	list_add(&var_table, "custom_packets", custom_packets, 0);
	list_add(&var_table, "config", config, 0);
	list_add(&var_table, "var_table", var_table, 0);

	list_add(&var_table, "sha1", sha1, 0);
	list_add(&var_table, "generate_session", generate_session, 0);
	list_add(&var_table, "check_session", check_session, 0);

	list_add(&var_table, "get_user", get_user, 0);
	list_add(&var_table, "put_user", put_user, 0);
	list_add(&var_table, "del_user", del_user, 0);

	list_add(&var_table, "list_add", list_add, 0);
	list_add(&var_table, "list_get", list_get, 0);
	list_add(&var_table, "list_del", list_del, 0);

	list_add(&var_table, "logger", logger, 0);
	list_add(&var_table, "str_event", str_event, 0);
	list_add(&var_table, "send_file", MC_Sendfile, 0);
	list_add(&var_table, "recv_file", MC_Recvfile, 0);

	logger(LOG_DEBUG, "Masterserver's initialization finished");

	char *daemonize = config_get("dbg_daemonize");
	if (daemonize == NULL || strcmp(daemonize, "1") == 0)
	{
		/* `man daemon` */
		if (daemon(1, 0) == -1)
		{
			logger(LOG_CRIT, "Can't daemonize! I'll keep running anyway...");
		}
		else
		{
			logger(LOG_DEBUG, "Daemonization went OK");
		}
	}
	logger(LOG_INFO, "Minecraft Masterserver 0.0 started normally");

}

/*
*  TODO: by now there's almost no cleanup at all...
*  Masterserver should work forever, but anyway.
*/
void MC_Cleanup(void)
{
	int ret;

	shutdown(global.socket, SHUT_RDWR);
	close(global.socket);

	ret = db_cleanup();
	if (ret != 0)
	{
		logger(LOG_ERROR, "db_cleanup() returned %d", ret);
	}

	logger(LOG_DEBUG, "Masterserver shut down completely");
}

int MC_RunThread(void *(*func)(void*), void *param)
{
	pthread_t thread;
	pthread_attr_t attr;
	pthread_attr_init(&attr);

	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_attr_setstacksize(&attr, 131072);

	if (pthread_create(&thread, &attr, func, param) != 0)
	{
		logger(LOG_ERROR, "Unable to create thread [func = %p, param = %p]", func, param);
		return -1;
	}

	pthread_attr_destroy(&attr);
	return 0;
}

void MC_InitModules(void)
{
	struct dirent *mod_dirent;
	DIR *modules_dir = opendir("modules");
	if (modules_dir == NULL)
	{
		logger(LOG_ERROR, "Couldn't open directory \"modules\"");
		return;
	}

	char *filename = NULL;
	char *ext = NULL;
	size_t namelen = 0;

	while((mod_dirent = readdir(modules_dir)) != NULL)
	{
		filename = mod_dirent->d_name;
		namelen = strlen(filename);
		if (namelen < 3)
		{
			continue;
		}

		if (strncmp(filename, "mod_", 4) != 0)
		{
			continue;
		}

		ext = filename + (namelen - 3);
		if (strcmp(ext, ".so") != 0)
		{
			continue;
		}
		logger(LOG_DEBUG, "Loading module \"%s\"", filename);
		MC_LoadModule(filename);
	}

	closedir(modules_dir);
}

void MC_LoadModule(char *mod_name)
{
	char *error = NULL;
	char mod_path[FILENAME_MAX+1];
	void *mod_handle = NULL;

	kv_list *module_ptr = list_getlast(modules);
	struct module_t curr_module;
	struct module_func_t *mod_func;
	struct module_t (*module_init)(void *(*config_get)(char*));

	sprintf(mod_path, "modules/%s", mod_name);
	mod_handle = dlopen(mod_path, RTLD_NOW);
	if (mod_handle == NULL)
	{
		logger(LOG_WARN, "dlopen(\"%s\"): %s", mod_name, dlerror());
		return;
	}

	module_init = dlsym(mod_handle, "module_init");
	if ((error = dlerror()) != NULL)
	{
		logger(LOG_WARN, "Couldn't find \"module_init\" symbol in \"%s\"", mod_name);
		dlclose(mod_handle);
		return;
	}

	/* we need module_ptr as we must 'bind' a module to a certain packet */
	curr_module = module_init(config_get);
	for (int i=0; i<curr_module.func_num; i++)
	{
		mod_func = curr_module.mod_func;
		if (modules == NULL)
		{
			list_add(&modules, mod_func[i].name, &mod_func[i], 0);
			module_ptr = modules;
		}
		else
		{
			module_ptr = list_append(&module_ptr, mod_func[i].name, &mod_func[i], 0);
		}

		if ((mod_func[i].event == CUSTOM_PACKET) && (mod_func[i].param != NULL))
		{
			list_add(&custom_packets, mod_func[i].param, module_ptr->value, 0);
		}
	}
}

int MC_RunModules(int event, void *param)
{
	if (modules == NULL) return -1;
	void **func_params = NULL;

	switch (event)
	{
	/* MASTER_START, MASTER_STOP, NEW_CONNECTION [same action] */
	case MASTER_START:
	case MASTER_STOP:
	case NEW_CONNECTION:
	{
		kv_list *list = modules;
		struct module_func_t *func = NULL;
		while (1)
		{
			func = list->value;

			if (func->event == event)
			{
				if (func->func_params != NULL)
				{
					func_params = list_get_arr(func->func_params);
				}
				func->addr(func_params);
				free(func_params);
			}
			if (list->next == NULL)
			{
				break;
			}
			list = list->next;
		}
		break;
	}
	case CUSTOM_PACKET:
	{
		if (param == NULL)
		{
			return -1;
		}

		char *match = param;
		kv_list *packet = custom_packets;
		struct module_func_t *func = NULL;

		while (1)
		{
			if (strncmp(match, packet->key, strlen(packet->key)) == 0)
			{
				func = packet->value;
				if (func->func_params != NULL)
				{
					func_params = list_get_arr(func->func_params);
				}
				func->addr(func_params);
				free(func_params);
			}
			if (packet->next == NULL)
			{
				break;
			}
			packet = packet->next;
		}
		break;
	}
	default:
		break;
	}
	return 0;
}

void signal_handler(int sig)
{
	switch (sig)
	{
	/* SIGINT, SIGTERM [same action] */
	case SIGINT:
	case SIGTERM:
	{
		logger(LOG_DEBUG, "Received %s, cleaning up", signal_names[sig]);
		MC_Cleanup();
		exit(EXIT_SUCCESS);
	}
	default:
		return;
	}
}

/* TODO: eh.. fix smth */
void init_database_lib(char *lib, int (**init)(void), int (**cleanup)(void), int (**get)(char*, struct usr_t *), int (**put)(char*, struct usr_t *), int (**del)(char*))
{
	char *error = NULL;
	void *init_t, *cleanup_t, *get_t, *put_t, *del_t;

	size_t lib_len = strlen(lib);
	char lib_path[lib_len + strlen("modules/") + 1];
	sprintf(lib_path, "modules/%s", lib);

	logger(LOG_DEBUG, "Loading database module: \"%s\"", lib);

	void *handle = dlopen(lib_path, RTLD_NOW);
	if (handle == NULL)
	{
		logger(LOG_WARN, "Unable to open database lib: %s", dlerror());
		return;
	}

	init_t = dlsym(handle, "db_init");
	if ((error = dlerror()) != NULL)
	{
		logger(LOG_WARN, "Unable to find \"db_init\" symbol in \"%s\"", lib);
		dlclose(handle);
		return;
	}

	cleanup_t = dlsym(handle, "db_cleanup");
	if ((error = dlerror()) != NULL)
        {
                logger(LOG_WARN, "Unable to find \"db_cleanup\" symbol in \"%s\"", lib);
                dlclose(handle);
                return;
        }

	get_t = dlsym(handle, "get_user");
	if ((error = dlerror()) != NULL)
        {
                logger(LOG_WARN, "Unable to find \"get_user\" symbol in \"%s\"", lib);
                dlclose(handle);
                return;
        }

	put_t = dlsym(handle, "put_user");
	if ((error = dlerror()) != NULL)
        {
                logger(LOG_WARN, "Unable to find \"put_user\" symbol in \"%s\"", lib);
                dlclose(handle);
                return;
        }

	del_t = dlsym(handle, "del_user");
	if ((error = dlerror()) != NULL)
        {
                logger(LOG_WARN, "Unable to find \"del_user\" symbol in \"%s\"", lib);
                dlclose(handle);
                return;
        }

	/* set functions only if ALL external functions are present */
	*init = init_t;
	*cleanup = cleanup_t;
	*get = get_t;
	*put = put_t;
	*del = del_t;
}

/* lol, php */
char *strtolower(char *dest, const char *src)
{
	int i;
	for (i=0; src[i] != '\0'; i++)
	{
		if (src[i] >= 'A' && src[i] <= 'Z')
		{
			dest[i] = src[i] + 'a' - 'A';
		}
		else
		{
			dest[i] = src[i];
		}
	}
	dest[i] = '\0';
	return dest;
}