#include <stdlib.h>
#include <string.h>
#include "kv_list.h"

/*
*  if `flags` contains FL_FREE, the `value` will be free()'d after deleting or replacing
*  replacing means adding a node with non-unique `key` into the list
*/
int list_add (kv_list **list, char *key, void *value, char flags)
{
	kv_list *tmp = NULL;
	kv_list *found = NULL;

	if (*list == NULL)
	{
		*list = malloc(sizeof(kv_list));
		tmp = *list;
	}
	else
	{
		tmp = *list;
		while(1)
		{
			if (strcmp(tmp->key, key) == 0)
			{
				found = tmp;
				break;
			}
			if (tmp->next == NULL) break;
			tmp = tmp->next;
		}
		if (found == NULL)
		{
			tmp->next = malloc(sizeof(kv_list));
			tmp = tmp->next;
		}
	}

	if ((found != NULL) && (tmp->flags & FL_FREE))
	{
		free(tmp->value);
	}

	tmp->value = value;
	tmp->flags = flags;

	if (found == NULL)
	{
		size_t key_size = strlen(key) + 1;

		tmp->key = malloc(key_size);
		memcpy(tmp->key, key, key_size);

		tmp->next = NULL;
	}

	return 0;
}

/*
*  it's very fast when you're adding a bunch of consequent nodes
*  because there's no need to go through the whole list after every insertion
*/
kv_list *list_append (kv_list **list, char *key, void *value, char flags)
{
	size_t key_size;
	kv_list *tmp = NULL;

	if (*list == NULL)
	{
		*list = malloc(sizeof(kv_list));
		tmp = *list;
	}
	else
	{
		(*list)->next = malloc(sizeof(kv_list));
		tmp = (*list)->next;
	}

	key_size = strlen(key) + 1;

	tmp->key = malloc(key_size);
	memcpy(tmp->key, key, key_size);

	tmp->value = value;
	tmp->flags = flags;
	tmp->next = NULL;

	return tmp;
}

void *list_get (kv_list *list, char *key)
{
	while(1)
	{
		if (strcmp(list->key, key) == 0)
		{
			return list->value;
		}
		if (list->next == NULL) break;
		list = list->next;
	}
	return NULL;
}

kv_list* list_getlast (kv_list *list)
{
	if (list == NULL) return NULL;

	while(list->next != NULL)
	{
		list = list->next;
	}

	return list;
}

int list_del (kv_list **list, char *key)
{
	kv_list *prev = NULL;
	kv_list *tmp = NULL;

	/* check the first node separately, so we could replace it */
	if (strcmp((*list)->key, key) == 0)
	{
		free((*list)->key);
		if ((*list)->flags & FL_FREE)
		{
			free((*list)->value);
		}

		prev = *list;
		*list = (*list)->next;

		free(prev);
		return 0;
	}

	tmp = (*list)->next;
	prev = *list;
	while(1)
	{
		if (strcmp(tmp->key, key) == 0)
		{
			free(tmp->key);
			if (tmp->flags & FL_FREE)
			{
				free(tmp->value);
			}

			prev->next = tmp->next;
			free(tmp);
			return 0;
		}
		if (tmp->next == NULL) break;
		prev = tmp;
		tmp = tmp->next;
	}

	return 0;
}

int list_destroy (kv_list **list)
{
	kv_list *tmp = *list;
	kv_list *next = NULL;

	if (tmp == NULL) return 0;

	while(1)
	{
		next = tmp->next;

		free(tmp->key);
		if (tmp->flags & FL_FREE)
		{
			free(tmp->value);
		}
		free(tmp);

		if (next == NULL) break;
		tmp = next;
	}
	*list = NULL;
	return 0;
}

size_t list_count (kv_list *list)
{
	size_t counter = 0;
	if (list == NULL) return counter;

	counter++;
	while(list->next != NULL)
	{
		counter++;
		list = list->next;
	}

	return counter;
}