#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/sendfile.h>
#include <string.h>
#include <unistd.h>

#include "datatypes.h"
#include "networking.h"
#include "logger.h"

int MC_PrepareSocket(void)
{
	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0)
	{
		logger(LOG_CRIT, "Unable to create listening socket");
		return -1;
	}
	else
	{
		logger(LOG_DEBUG, "Socket created: %d", sock);
	}

	int reuse = 1;
	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuse, sizeof(int)) < 0)
	{
		logger(LOG_CRIT, "Unable to set SO_REUSEADDR on socket %d", sock);
		close(sock);
		return -1;
	}

	/* Getting server port from config */
	uint16_t server_port = SERV_PORT;
	char *str_port = config_get("server_port");
	if (str_port != NULL)
	{
		server_port = strtol(str_port, NULL, 10);
	}

	/* Get server ip from config */
	char *str_ip = config_get("server_ip");
	if (str_ip == NULL)
	{
		str_ip = "0.0.0.0";
	}

	struct sockaddr_in serv_addr;
	serv_addr.sin_family = AF_INET;
	if (inet_aton(str_ip, &serv_addr.sin_addr) == 0)
	{
		logger(LOG_WARN, "server_ip (%s) is not valid, I will use 0.0.0.0", str_ip);
	}
	serv_addr.sin_port = htons(server_port);

	if (bind(sock, (struct sockaddr*)&serv_addr, sizeof(serv_addr)) < 0)
	{
		logger(LOG_CRIT, "Unable to bind to %s:%u", inet_ntoa(serv_addr.sin_addr), server_port);
		close(sock);
		return -1;
	}
	else
	{
		logger(LOG_DEBUG, "Bound to %s:%u", inet_ntoa(serv_addr.sin_addr), server_port);
	}

	if (listen(sock, SOMAXCONN) < 0)
	{
		logger(LOG_CRIT, "Unable to start listening on socket %d", sock);
		close(sock);
		return -1;
	}
	else
	{
		logger(LOG_DEBUG, "Listening for connections on %s:%u [sock: %d]", inet_ntoa(serv_addr.sin_addr), server_port, sock);
	}

	return sock;
}

void MC_TuneSocket(int sock)
{
	struct timeval tv;
	tv.tv_sec = 20;
	tv.tv_usec = 0;

	if (setsockopt(sock, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(struct timeval)) != 0)
	{
		logger(LOG_WARN, "Unable to set SO_RCVTIMEO on socket %d", sock);
	}

	if (setsockopt(sock, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(struct timeval)) != 0)
	{
		logger(LOG_WARN, "Unable to set SO_SNDTIMEO on socket %d", sock);
	}
}

ssize_t MC_RecvAll(int sock, char *buf, size_t max)
{
	int chunk = RECV_BUF;
	ssize_t recv_now = 0;
	ssize_t recv_total = 0;
	while ((recv_now = recv(sock, buf+recv_total, chunk, 0)) > 0)
	{
		recv_total += recv_now;

		if (recv_total + chunk > (ssize_t)max)
		{
			chunk = RECV_MAX - recv_total;
		}

		if ((recv_now < RECV_BUF) || (recv_total == (ssize_t)max))
		{
			break;
		}
	}

	return recv_total;
}

int MC_Recvfile(int sockfd, char *filename, size_t max_size)
{
	int filefd;
	char buf[8*1024];
	ssize_t bytes_recv;
	size_t total_recv = 0;
	int ret;

	filefd = open(filename, O_WRONLY | O_CREAT);
	if (filefd == -1)
	{
		perror("[!!!] open()");
		return -1;
	}

	while ((bytes_recv = recv(sockfd, buf, 8*1024, 0)) > 0)
	{
		if (total_recv + bytes_recv > max_size)
		{
			printf("[!!!] exceed!\n");
			close(filefd);
			return 0;
		}
		total_recv += bytes_recv;

		ret = write(filefd, buf, bytes_recv);
		if (ret == -1)
		{
			perror("[!!!] write()");
			close(filefd);
			return -1;
		}
	}

	close(filefd);
	return 0;
}

/*
*  Uses the most efficient way to send file to socket
*  If real_name is not NULL, an appropriate Content-Disposition header will be sent
*/
int MC_Sendfile(int sockfd, char *filename, char *real_name)
{
	int res;
	int filefd;
	struct stat fileinfo;
	char http_header[256] = { 0 };

	filefd = open(filename, O_RDONLY);
	if (filefd == -1)
	{
		return -1;
	}

	res = fstat(filefd, &fileinfo);
	if (res == -1)
	{
		return -1;
	}

	if (real_name != NULL)
	{
		sprintf(http_header, "HTTP/1.0 200 OK\nContent-Type: application/octet-stream\nContent-Disposition: attachment; filename=%s\nContent-Length: %u\n\n", real_name, (unsigned int)fileinfo.st_size);
	}
	else
	{
		sprintf(http_header, "HTTP/1.0 200 OK\nContent-Type: application/octet-stream\nContent-Length: %u\n\n", (unsigned int)fileinfo.st_size);
	}

	res = send(sockfd, http_header, strlen(http_header), MSG_MORE);
	if (res == -1)
	{
		/* Non-critical */
	}

	res = sendfile(sockfd, filefd, NULL, fileinfo.st_size);
	if (res == -1)
	{
		close(filefd);
		return -1;
	}

	close(filefd);
	return 0;
}

ssize_t sendstr(int sockfd, const char *str)
{
	return send(sockfd, str, strlen(str), 0);
}